@@ -9,21 +9,22 @@ import (
 	"reflect"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 
-	"k8s.io/klog"
-
 	honode "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/controller"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/cni"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/factory"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/informer"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/kube"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
-
 	kapi "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/client-go/kubernetes"
 	"k8s.io/client-go/tools/cache"
+	"k8s.io/client-go/tools/record"
+	"k8s.io/klog"
 )
 
 // OvnNode is the object holder for utilities meant for node management
@@ -32,35 +33,32 @@ type OvnNode struct {
 	Kube         kube.Interface
 	watchFactory *factory.WatchFactory
 	stopChan     chan struct{}
+	recorder     record.EventRecorder
 }
 
 // NewNode creates a new controller for node management
-func NewNode(kubeClient kubernetes.Interface, wf *factory.WatchFactory, name string, stopChan chan struct{}) *OvnNode {
+func NewNode(kubeClient kubernetes.Interface, wf *factory.WatchFactory, name string, stopChan chan struct{}, eventRecorder record.EventRecorder) *OvnNode {
 	return &OvnNode{
 		name:         name,
 		Kube:         &kube.Kube{KClient: kubeClient},
 		watchFactory: wf,
 		stopChan:     stopChan,
+		recorder:     eventRecorder,
 	}
 }
 
 func setupOVNNode(node *kapi.Node) error {
 	var err error
 
-	nodeName, err := util.GetNodeHostname(node)
-	if err != nil {
-		return fmt.Errorf("failed to obtain hostname from node %q: %v", node.Name, err)
-	}
-
-	nodeIP := config.Default.EncapIP
-	if nodeIP == "" {
-		nodeIP, err = util.GetNodeIP(node)
+	encapIP := config.Default.EncapIP
+	if encapIP == "" {
+		encapIP, err = util.GetNodePrimaryIP(node)
 		if err != nil {
 			return fmt.Errorf("failed to obtain local IP from node %q: %v", node.Name, err)
 		}
 	} else {
-		if ip := net.ParseIP(nodeIP); ip == nil {
-			return fmt.Errorf("invalid encapsulation IP provided %q", nodeIP)
+		if ip := net.ParseIP(encapIP); ip == nil {
+			return fmt.Errorf("invalid encapsulation IP provided %q", encapIP)
 		}
 	}
 
@@ -68,12 +66,12 @@ func setupOVNNode(node *kapi.Node) error {
 		"Open_vSwitch",
 		".",
 		fmt.Sprintf("external_ids:ovn-encap-type=%s", config.Default.EncapType),
-		fmt.Sprintf("external_ids:ovn-encap-ip=%s", nodeIP),
+		fmt.Sprintf("external_ids:ovn-encap-ip=%s", encapIP),
 		fmt.Sprintf("external_ids:ovn-remote-probe-interval=%d",
 			config.Default.InactivityProbe),
 		fmt.Sprintf("external_ids:ovn-openflow-probe-interval=%d",
 			config.Default.OpenFlowProbe),
-		fmt.Sprintf("external_ids:hostname=\"%s\"", nodeName),
+		fmt.Sprintf("external_ids:hostname=\"%s\"", node.Name),
 		"external_ids:ovn-monitor-all=true",
 	)
 	if err != nil {
@@ -115,7 +113,7 @@ func isOVNControllerReady(name string) (bool, error) {
 		ctlFile := runDir + fmt.Sprintf("ovn-controller.%s.ctl", strings.TrimSuffix(string(pid), "\n"))
 		ret, _, err := util.RunOVSAppctl("-t", ctlFile, "connection-status")
 		if err == nil {
-			klog.Infof("node %s connection status = %s", name, ret)
+			klog.Infof("Node %s connection status = %s", name, ret)
 			return ret == "connected", nil
 		}
 		return false, err
@@ -138,9 +136,16 @@ func isOVNControllerReady(name string) (bool, error) {
 	err = wait.PollImmediate(500*time.Millisecond, 60*time.Second, func() (bool, error) {
 		stdout, _, err := util.RunOVSOfctl("dump-aggregate", "br-int")
 		if err != nil {
+			klog.V(5).Infof("Error dumping aggregate flows: %v "+
+				"for node: %s", err, name)
 			return false, nil
 		}
-		return !strings.Contains(stdout, "flow_count=0"), nil
+		ret := strings.Contains(stdout, "flow_count=0")
+		if ret {
+			klog.V(5).Infof("Got a flow count of 0 when "+
+				"dumping flows for node: %s", name)
+		}
+		return !ret, nil
 	})
 	if err != nil {
 		return false, fmt.Errorf("timed out dumping br-int flow entries for node %s: %v", name, err)
@@ -149,9 +154,9 @@ func isOVNControllerReady(name string) (bool, error) {
 	return true, nil
 }
 
-// Start learns the subnet assigned to it by the master controller
+// Start learns the subnets assigned to it by the master controller
 // and calls the SetupNode script which establishes the logical switch
-func (n *OvnNode) Start() error {
+func (n *OvnNode) Start(wg *sync.WaitGroup) error {
 	var err error
 	var node *kapi.Node
 	var subnets []*net.IPNet
@@ -160,7 +165,7 @@ func (n *OvnNode) Start() error {
 	// Log level is returned to configured value when bring up is complete.
 	var level klog.Level
 	if err := level.Set("5"); err != nil {
-		klog.Errorf("setting klog \"loglevel\" to 5 failed, err: %v", err)
+		klog.Errorf("Setting klog \"loglevel\" to 5 failed, err: %v", err)
 	}
 
 	for _, auth := range []config.OvnAuthConfig{config.OvnNorth, config.OvnSouth} {
@@ -180,12 +185,12 @@ func (n *OvnNode) Start() error {
 	// First wait for the node logical switch to be created by the Master, timeout is 300s.
 	err = wait.PollImmediate(500*time.Millisecond, 300*time.Second, func() (bool, error) {
 		if node, err = n.Kube.GetNode(n.name); err != nil {
-			klog.Infof("waiting to retrieve node %s: %v", n.name, err)
+			klog.Infof("Waiting to retrieve node %s: %v", n.name, err)
 			return false, nil
 		}
 		subnets, err = util.ParseNodeHostSubnetAnnotation(node)
 		if err != nil {
-			klog.Infof("waiting for node %s to start, no annotation found on node for subnet: %v", n.name, err)
+			klog.Infof("Waiting for node %s to start, no annotation found on node for subnet: %v", n.name, err)
 			return false, nil
 		}
 		return true, nil
@@ -204,8 +209,7 @@ func (n *OvnNode) Start() error {
 	waiter := newStartupWaiter()
 
 	// Initialize gateway resources on the node
-	// FIXME DUAL-STACK
-	if err := n.initGateway(subnets[0], nodeAnnotator, waiter); err != nil {
+	if err := n.initGateway(subnets, nodeAnnotator, waiter); err != nil {
 		return err
 	}
 
@@ -215,7 +219,7 @@ func (n *OvnNode) Start() error {
 	}
 
 	if err := nodeAnnotator.Run(); err != nil {
-		return fmt.Errorf("Failed to set node %s annotations: %v", n.name, err)
+		return fmt.Errorf("failed to set node %s annotations: %v", n.name, err)
 	}
 
 	// Wait for management port and gateway resources to be created by the master
@@ -227,13 +231,26 @@ func (n *OvnNode) Start() error {
 	klog.Infof("Gateway and management port readiness took %v", time.Since(start))
 
 	if config.HybridOverlay.Enabled {
-		if err := honode.StartNode(n.name, n.Kube, n.watchFactory, n.stopChan); err != nil {
+		factory := n.watchFactory.GetFactory()
+		nodeController, err := honode.NewNode(
+			n.Kube,
+			n.name,
+			factory.Core().V1().Nodes().Informer(),
+			factory.Core().V1().Pods().Informer(),
+			informer.NewDefaultEventHandler,
+		)
+		if err != nil {
 			return err
 		}
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			nodeController.Run(n.stopChan)
+		}()
 	}
 
 	if err := level.Set(strconv.Itoa(config.Logging.Level)); err != nil {
-		klog.Errorf("reset of initial klog \"loglevel\" failed, err: %v", err)
+		klog.Errorf("Reset of initial klog \"loglevel\" failed, err: %v", err)
 	}
 
 	// start health check to ensure there are no stale OVS internal ports
@@ -250,12 +267,9 @@ func (n *OvnNode) Start() error {
 
 	kclient, ok := n.Kube.(*kube.Kube)
 	if !ok {
-		return fmt.Errorf("Cannot get kubeclient for starting CNI server")
-	}
-	err = n.WatchEndpoints()
-	if err != nil {
-		return err
+		return fmt.Errorf("cannot get kubeclient for starting CNI server")
 	}
+	n.WatchEndpoints()
 
 	// start the cni server
 	cniServer := cni.NewCNIServer("", kclient.KClient)
@@ -264,8 +278,8 @@ func (n *OvnNode) Start() error {
 	return err
 }
 
-func (n *OvnNode) WatchEndpoints() error {
-	_, err := n.watchFactory.AddEndpointsHandler(cache.ResourceEventHandlerFuncs{
+func (n *OvnNode) WatchEndpoints() {
+	n.watchFactory.AddEndpointsHandler(cache.ResourceEventHandlerFuncs{
 		UpdateFunc: func(old, new interface{}) {
 			epNew := new.(*kapi.Endpoints)
 			epOld := old.(*kapi.Endpoints)
@@ -293,7 +307,6 @@ func (n *OvnNode) WatchEndpoints() error {
 			}
 		},
 	}, nil)
-	return err
 }
 
 type epAddressItem struct {
