@@ -12,6 +12,13 @@ import (
 )
 
 var _ = Describe("Gateway Init Operations", func() {
+	BeforeEach(func() {
+		// Restore global default values before each testcase
+		config.PrepareTestConfig()
+		// TODO make contexts here for shared gw mode and local gw mode, right now this only tests shared gw
+		config.Gateway.Mode = config.GatewayModeShared
+	})
+
 	It("correctly sorts gateway routers", func() {
 		fexec := ovntest.NewFakeExec()
 		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
@@ -24,12 +31,6 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 
 		err := util.SetExec(fexec)
 		Expect(err).NotTo(HaveOccurred())
-
-		name, ip, err := getDefaultGatewayRouterIP()
-		Expect(err).NotTo(HaveOccurred())
-		Expect(name).To(Equal("node1"))
-		Expect(ip.String()).To(Equal("100.64.0.1"))
-		Expect(fexec.CalledMatchesExpected()).To(BeTrue(), fexec.ErrorDesc)
 	})
 
 	It("ignores malformatted gateway router entires", func() {
@@ -44,26 +45,20 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 
 		err := util.SetExec(fexec)
 		Expect(err).NotTo(HaveOccurred())
-
-		name, ip, err := getDefaultGatewayRouterIP()
-		Expect(err).NotTo(HaveOccurred())
-		Expect(name).To(Equal("node4"))
-		Expect(ip.String()).To(Equal("100.64.0.4"))
-		Expect(fexec.CalledMatchesExpected()).To(BeTrue(), fexec.ErrorDesc)
 	})
 
 	It("creates an IPv4 gateway in OVN", func() {
-		clusterIPSubnet := []*net.IPNet{ovntest.MustParseIPNet("10.128.0.0/14")}
-		hostSubnets := []*net.IPNet{ovntest.MustParseIPNet("10.130.0.0/23")}
-		joinSubnets := []*net.IPNet{ovntest.MustParseIPNet("100.64.0.0/29")}
+		clusterIPSubnets := ovntest.MustParseIPNets("10.128.0.0/14")
+		hostSubnets := ovntest.MustParseIPNets("10.130.0.0/23")
+		joinSubnets := ovntest.MustParseIPNets("100.64.0.0/29")
 		nodeName := "test-node"
 		l3GatewayConfig := &util.L3GatewayConfig{
 			Mode:           config.GatewayModeLocal,
 			ChassisID:      "SYSTEM-ID",
 			InterfaceID:    "INTERFACE-ID",
 			MACAddress:     ovntest.MustParseMAC("11:22:33:44:55:66"),
-			IPAddresses:    []*net.IPNet{ovntest.MustParseIPNet("169.254.33.2/24")},
-			NextHops:       []net.IP{ovntest.MustParseIP("169.254.33.1")},
+			IPAddresses:    ovntest.MustParseIPNets("169.254.33.2/24"),
+			NextHops:       ovntest.MustParseIPs("169.254.33.1"),
 			NodePortEnable: true,
 		}
 		sctpSupport := false
@@ -101,6 +96,9 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		})
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 set logical_router GR_test-node load_balancer=" + tcpLBUUID + "," + udpLBUUID,
+			"ovn-nbctl --timeout=15 get logical_switch test-node load_balancer",
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + tcpLBUUID,
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + udpLBUUID,
 		})
 
 		fexec.AddFakeCmdsNoOutputNoError([]string{
@@ -110,26 +108,26 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 			"ovn-nbctl --timeout=15 -- --may-exist lsp-add ext_test-node etor-GR_test-node -- set logical_switch_port etor-GR_test-node type=router options:router-port=rtoe-GR_test-node addresses=\"11:22:33:44:55:66\"",
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node 0.0.0.0/0 169.254.33.1 rtoe-GR_test-node",
 			"ovn-nbctl --timeout=15 --may-exist --policy=src-ip lr-route-add ovn_cluster_router 10.130.0.0/23 100.64.0.1",
-			"ovn-nbctl --timeout=15 --may-exist lr-nat-add GR_test-node snat 169.254.33.2 10.128.0.0/14",
+			"ovn-nbctl --timeout=15 --if-exists lr-nat-del GR_test-node snat 10.128.0.0/14 -- lr-nat-add GR_test-node snat 169.254.33.2 10.128.0.0/14",
 		})
 
-		err = gatewayInit(nodeName, clusterIPSubnet, hostSubnets, joinSubnets, l3GatewayConfig, sctpSupport)
+		err = gatewayInit(nodeName, clusterIPSubnets, hostSubnets, joinSubnets, l3GatewayConfig, sctpSupport)
 		Expect(err).NotTo(HaveOccurred())
 		Expect(fexec.CalledMatchesExpected()).To(BeTrue())
 	})
 
 	It("creates an IPv6 gateway in OVN", func() {
-		clusterIPSubnet := []*net.IPNet{ovntest.MustParseIPNet("fd01::/48")}
-		hostSubnets := []*net.IPNet{ovntest.MustParseIPNet("fd01:0:0:2::/64")}
-		joinSubnets := []*net.IPNet{ovntest.MustParseIPNet("fd98::/125")}
+		clusterIPSubnets := ovntest.MustParseIPNets("fd01::/48")
+		hostSubnets := ovntest.MustParseIPNets("fd01:0:0:2::/64")
+		joinSubnets := ovntest.MustParseIPNets("fd98::/125")
 		nodeName := "test-node"
 		l3GatewayConfig := &util.L3GatewayConfig{
 			Mode:           config.GatewayModeLocal,
 			ChassisID:      "SYSTEM-ID",
 			InterfaceID:    "INTERFACE-ID",
 			MACAddress:     ovntest.MustParseMAC("11:22:33:44:55:66"),
-			IPAddresses:    []*net.IPNet{ovntest.MustParseIPNet("fd99::2/64")},
-			NextHops:       []net.IP{ovntest.MustParseIP("fd99::1")},
+			IPAddresses:    ovntest.MustParseIPNets("fd99::2/64"),
+			NextHops:       ovntest.MustParseIPs("fd99::1"),
 			NodePortEnable: true,
 		}
 		sctpSupport := false
@@ -138,13 +136,15 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		err := util.SetExec(fexec)
 		Expect(err).NotTo(HaveOccurred())
 
+		// 0a:58:ee:33:fc:1a generated from util.IPAddrToHWAddr(net.ParseIP("fd98::1")).String()
+		// 0a:58:5f:8a:48:8c generated from util.IPAddrToHWAddr(net.ParseIP("fd98::2")).String()
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 -- --may-exist lr-add GR_test-node -- set logical_router GR_test-node options:chassis=SYSTEM-ID external_ids:physical_ip=fd99::2 external_ids:physical_ips=fd99::2",
 			"ovn-nbctl --timeout=15 -- --may-exist ls-add join_test-node",
 			"ovn-nbctl --timeout=15 -- --may-exist lsp-add join_test-node jtor-GR_test-node -- set logical_switch_port jtor-GR_test-node type=router options:router-port=rtoj-GR_test-node addresses=router",
-			"ovn-nbctl --timeout=15 -- --if-exists lrp-del rtoj-GR_test-node -- lrp-add GR_test-node rtoj-GR_test-node 0a:58:fd:98:00:01 fd98::1/125",
+			"ovn-nbctl --timeout=15 -- --if-exists lrp-del rtoj-GR_test-node -- lrp-add GR_test-node rtoj-GR_test-node 0a:58:ee:33:fc:1a fd98::1/125",
 			"ovn-nbctl --timeout=15 -- --may-exist lsp-add join_test-node jtod-test-node -- set logical_switch_port jtod-test-node type=router options:router-port=dtoj-test-node addresses=router",
-			"ovn-nbctl --timeout=15 -- --if-exists lrp-del dtoj-test-node -- lrp-add ovn_cluster_router dtoj-test-node 0a:58:fd:98:00:02 fd98::2/125",
+			"ovn-nbctl --timeout=15 -- --if-exists lrp-del dtoj-test-node -- lrp-add ovn_cluster_router dtoj-test-node 0a:58:5f:8a:48:8c fd98::2/125",
 			"ovn-nbctl --timeout=15 set logical_router GR_test-node options:lb_force_snat_ip=fd98::1",
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node fd01::/48 fd98::2",
 		})
@@ -167,6 +167,9 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		})
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 set logical_router GR_test-node load_balancer=" + tcpLBUUID + "," + udpLBUUID,
+			"ovn-nbctl --timeout=15 get logical_switch test-node load_balancer",
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + tcpLBUUID,
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + udpLBUUID,
 		})
 
 		fexec.AddFakeCmdsNoOutputNoError([]string{
@@ -176,41 +179,26 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 			"ovn-nbctl --timeout=15 -- --may-exist lsp-add ext_test-node etor-GR_test-node -- set logical_switch_port etor-GR_test-node type=router options:router-port=rtoe-GR_test-node addresses=\"11:22:33:44:55:66\"",
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node ::/0 fd99::1 rtoe-GR_test-node",
 			"ovn-nbctl --timeout=15 --may-exist --policy=src-ip lr-route-add ovn_cluster_router fd01:0:0:2::/64 fd98::1",
-			"ovn-nbctl --timeout=15 --may-exist lr-nat-add GR_test-node snat fd99::2 fd01::/48",
+			"ovn-nbctl --timeout=15 --if-exists lr-nat-del GR_test-node snat fd01::/48 -- lr-nat-add GR_test-node snat fd99::2 fd01::/48",
 		})
 
-		err = gatewayInit(nodeName, clusterIPSubnet, hostSubnets, joinSubnets, l3GatewayConfig, sctpSupport)
+		err = gatewayInit(nodeName, clusterIPSubnets, hostSubnets, joinSubnets, l3GatewayConfig, sctpSupport)
 		Expect(err).NotTo(HaveOccurred())
 		Expect(fexec.CalledMatchesExpected()).To(BeTrue())
 	})
 
 	It("creates a dual-stack gateway in OVN", func() {
-		clusterIPSubnets := []*net.IPNet{
-			ovntest.MustParseIPNet("10.128.0.0/14"),
-			ovntest.MustParseIPNet("fd01::/48"),
-		}
-		hostSubnets := []*net.IPNet{
-			ovntest.MustParseIPNet("10.130.0.0/23"),
-			ovntest.MustParseIPNet("fd01:0:0:2::/64"),
-		}
-		joinSubnets := []*net.IPNet{
-			ovntest.MustParseIPNet("100.64.0.0/29"),
-			ovntest.MustParseIPNet("fd98::/125"),
-		}
+		clusterIPSubnets := ovntest.MustParseIPNets("10.128.0.0/14", "fd01::/48")
+		hostSubnets := ovntest.MustParseIPNets("10.130.0.0/23", "fd01:0:0:2::/64")
+		joinSubnets := ovntest.MustParseIPNets("100.64.0.0/29", "fd98::/125")
 		nodeName := "test-node"
 		l3GatewayConfig := &util.L3GatewayConfig{
-			Mode:        config.GatewayModeLocal,
-			ChassisID:   "SYSTEM-ID",
-			InterfaceID: "INTERFACE-ID",
-			MACAddress:  ovntest.MustParseMAC("11:22:33:44:55:66"),
-			IPAddresses: []*net.IPNet{
-				ovntest.MustParseIPNet("169.254.33.2/24"),
-				ovntest.MustParseIPNet("fd99::2/64"),
-			},
-			NextHops: []net.IP{
-				ovntest.MustParseIP("169.254.33.1"),
-				ovntest.MustParseIP("fd99::1"),
-			},
+			Mode:           config.GatewayModeLocal,
+			ChassisID:      "SYSTEM-ID",
+			InterfaceID:    "INTERFACE-ID",
+			MACAddress:     ovntest.MustParseMAC("11:22:33:44:55:66"),
+			IPAddresses:    ovntest.MustParseIPNets("169.254.33.2/24", "fd99::2/64"),
+			NextHops:       ovntest.MustParseIPs("169.254.33.1", "fd99::1"),
 			NodePortEnable: true,
 		}
 		sctpSupport := false
@@ -226,7 +214,7 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 			"ovn-nbctl --timeout=15 -- --if-exists lrp-del rtoj-GR_test-node -- lrp-add GR_test-node rtoj-GR_test-node 0a:58:64:40:00:01 100.64.0.1/29 fd98::1/125",
 			"ovn-nbctl --timeout=15 -- --may-exist lsp-add join_test-node jtod-test-node -- set logical_switch_port jtod-test-node type=router options:router-port=dtoj-test-node addresses=router",
 			"ovn-nbctl --timeout=15 -- --if-exists lrp-del dtoj-test-node -- lrp-add ovn_cluster_router dtoj-test-node 0a:58:64:40:00:02 100.64.0.2/29 fd98::2/125",
-			"ovn-nbctl --timeout=15 set logical_router GR_test-node options:lb_force_snat_ip=100.64.0.1",
+			"ovn-nbctl --timeout=15 set logical_router GR_test-node options:lb_force_snat_ip=100.64.0.1 fd98::1",
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node 10.128.0.0/14 100.64.0.2",
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node fd01::/48 fd98::2",
 		})
@@ -249,6 +237,9 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		})
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 set logical_router GR_test-node load_balancer=" + tcpLBUUID + "," + udpLBUUID,
+			"ovn-nbctl --timeout=15 get logical_switch test-node load_balancer",
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + tcpLBUUID,
+			"ovn-nbctl --timeout=15 ls-lb-add test-node " + udpLBUUID,
 		})
 
 		fexec.AddFakeCmdsNoOutputNoError([]string{
@@ -260,8 +251,8 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 			"ovn-nbctl --timeout=15 --may-exist lr-route-add GR_test-node ::/0 fd99::1 rtoe-GR_test-node",
 			"ovn-nbctl --timeout=15 --may-exist --policy=src-ip lr-route-add ovn_cluster_router 10.130.0.0/23 100.64.0.1",
 			"ovn-nbctl --timeout=15 --may-exist --policy=src-ip lr-route-add ovn_cluster_router fd01:0:0:2::/64 fd98::1",
-			"ovn-nbctl --timeout=15 --may-exist lr-nat-add GR_test-node snat 169.254.33.2 10.128.0.0/14",
-			"ovn-nbctl --timeout=15 --may-exist lr-nat-add GR_test-node snat fd99::2 fd01::/48",
+			"ovn-nbctl --timeout=15 --if-exists lr-nat-del GR_test-node snat 10.128.0.0/14 -- lr-nat-add GR_test-node snat 169.254.33.2 10.128.0.0/14",
+			"ovn-nbctl --timeout=15 --if-exists lr-nat-del GR_test-node snat fd01::/48 -- lr-nat-add GR_test-node snat fd99::2 fd01::/48",
 		})
 
 		err = gatewayInit(nodeName, clusterIPSubnets, hostSubnets, joinSubnets, l3GatewayConfig, sctpSupport)
@@ -273,9 +264,8 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		nodeName := "test-node"
 		hostSubnet := ovntest.MustParseIPNet("10.130.0.0/23")
 		const (
-			nodeRouteUUID    string = "0cac12cf-3e0f-4682-b028-5ea2e0001962"
-			nodemgtRouteUUID string = "0cac12cf-3e0f-4682-b028-5ea2e0001963"
-			tcpLBUUID        string = "1a3dfc82-2749-4931-9190-c30e7c0ecea3"
+			nodeRouteUUID string = "0cac12cf-3e0f-4682-b028-5ea2e0001962"
+			tcpLBUUID     string = "1a3dfc82-2749-4931-9190-c30e7c0ecea3"
 		)
 
 		fexec := ovntest.NewFakeExec()
@@ -293,12 +283,7 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 --if-exists remove logical_router ovn_cluster_router static_routes " + nodeRouteUUID,
 		})
-		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-			Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=_uuid find logical_router_static_route nexthop=\"10.130.0.2\"",
-			Output: nodemgtRouteUUID,
-		})
 		fexec.AddFakeCmdsNoOutputNoError([]string{
-			"ovn-nbctl --timeout=15 --if-exists remove logical_router ovn_cluster_router static_routes " + nodemgtRouteUUID,
 			"ovn-nbctl --timeout=15 --if-exist ls-del join_test-node",
 			"ovn-nbctl --timeout=15 --if-exist lr-del GR_test-node",
 			"ovn-nbctl --timeout=15 --if-exist ls-del ext_test-node",
@@ -320,18 +305,16 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 lb-del " + tcpLBUUID,
 		})
+		cleanupPBRandNATRules(fexec, nodeName, []*net.IPNet{hostSubnet})
 
-		err = gatewayCleanup(nodeName, []*net.IPNet{hostSubnet})
+		err = gatewayCleanup(nodeName)
 		Expect(err).NotTo(HaveOccurred())
 		Expect(fexec.CalledMatchesExpected()).To(BeTrue())
 	})
 
 	It("cleans up a dual-stack gateway in OVN", func() {
 		nodeName := "test-node"
-		hostSubnets := []*net.IPNet{
-			ovntest.MustParseIPNet("10.130.0.0/23"),
-			ovntest.MustParseIPNet("fd01:0:0:2::/64"),
-		}
+		hostSubnets := ovntest.MustParseIPNets("10.130.0.0/23", "fd01:0:0:2::/64")
 		const (
 			v4RouteUUID    string = "0cac12cf-3e0f-4682-b028-5ea2e0001962"
 			v6RouteUUID    string = "0cac12cf-4682-3e0f-b028-5ea2e0001962"
@@ -362,20 +345,6 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 --if-exists remove logical_router ovn_cluster_router static_routes " + v6RouteUUID,
 		})
-		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-			Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=_uuid find logical_router_static_route nexthop=\"10.130.0.2\"",
-			Output: v4mgtRouteUUID,
-		})
-		fexec.AddFakeCmdsNoOutputNoError([]string{
-			"ovn-nbctl --timeout=15 --if-exists remove logical_router ovn_cluster_router static_routes " + v4mgtRouteUUID,
-		})
-		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-			Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=_uuid find logical_router_static_route nexthop=\"fd01:0:0:2::2\"",
-			Output: v6mgtRouteUUID,
-		})
-		fexec.AddFakeCmdsNoOutputNoError([]string{
-			"ovn-nbctl --timeout=15 --if-exists remove logical_router ovn_cluster_router static_routes " + v6mgtRouteUUID,
-		})
 
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 --if-exist ls-del join_test-node",
@@ -399,8 +368,9 @@ node4 chassis=912d592c-904c-40cd-9ef1-c2e5b49a33dd lb_force_snat_ip=100.64.0.4`,
 		fexec.AddFakeCmdsNoOutputNoError([]string{
 			"ovn-nbctl --timeout=15 lb-del " + tcpLBUUID,
 		})
+		cleanupPBRandNATRules(fexec, nodeName, hostSubnets)
 
-		err = gatewayCleanup(nodeName, hostSubnets)
+		err = gatewayCleanup(nodeName)
 		Expect(err).NotTo(HaveOccurred())
 		Expect(fexec.CalledMatchesExpected()).To(BeTrue())
 	})
