@@ -2,10 +2,14 @@ package ovn
 
 import (
 	"fmt"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 	"net"
+	"strings"
 	"time"
 
 	hotypes "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/types"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
+
 	kapi "k8s.io/api/core/v1"
 	utilwait "k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/klog"
@@ -13,7 +17,10 @@ import (
 
 const (
 	// Annotation used to enable/disable multicast in the namespace
-	nsMulticastAnnotation = "k8s.ovn.org/multicast-enabled"
+	nsMulticastAnnotation        = "k8s.ovn.org/multicast-enabled"
+	routingExternalGWsAnnotation = "k8s.ovn.org/routing-external-gws"
+	routingNamespaceAnnotation   = "k8s.ovn.org/routing-namespaces"
+	routingNetworkAnnotation     = "k8s.ovn.org/routing-network"
 )
 
 func (oc *Controller) syncNamespaces(namespaces []interface{}) {
@@ -27,11 +34,11 @@ func (oc *Controller) syncNamespaces(namespaces []interface{}) {
 		expectedNs[ns.Name] = true
 	}
 
-	err := oc.forEachAddressSetUnhashedName(func(addrSetName,
-		namespaceName, nameSuffix string) {
+	err := oc.addressSetFactory.ForEachAddressSet(func(addrSetName, namespaceName, nameSuffix string) {
 		if nameSuffix == "" && !expectedNs[namespaceName] {
-			// delete the address sets for this namespace from OVN
-			deleteAddressSet(hashedAddressSet(addrSetName))
+			if err := oc.addressSetFactory.DestroyAddressSetInBackingStore(addrSetName); err != nil {
+				klog.Errorf(err.Error())
+			}
 		}
 	})
 	if err != nil {
@@ -40,21 +47,16 @@ func (oc *Controller) syncNamespaces(namespaces []interface{}) {
 }
 
 func (oc *Controller) addPodToNamespace(ns string, portInfo *lpInfo) error {
-	nsInfo := oc.getNamespaceLocked(ns)
-	if nsInfo == nil {
-		return nil
+	nsInfo, err := oc.waitForNamespaceLocked(ns)
+	if err != nil {
+		return err
 	}
 	defer nsInfo.Unlock()
 
-	// If pod has already been added, nothing to do.
-	address := portInfo.ip.String()
-	if nsInfo.addressSet[address] != "" {
-		return nil
+	if err := nsInfo.addressSet.AddIPs(createIPAddressSlice(portInfo.ips)); err != nil {
+		return err
 	}
 
-	nsInfo.addressSet[address] = portInfo.name
-	addToAddressSet(hashedAddressSet(ns), address)
-
 	// If multicast is allowed and enabled for the namespace, add the port
 	// to the allow policy.
 	if oc.multicastSupport && nsInfo.multicastEnabled {
@@ -73,14 +75,10 @@ func (oc *Controller) deletePodFromNamespace(ns string, portInfo *lpInfo) error
 	}
 	defer nsInfo.Unlock()
 
-	address := portInfo.ip.String()
-	if nsInfo.addressSet[address] == "" {
-		return nil
+	if err := nsInfo.addressSet.DeleteIPs(createIPAddressSlice(portInfo.ips)); err != nil {
+		return err
 	}
 
-	delete(nsInfo.addressSet, address)
-	removeFromAddressSet(hashedAddressSet(ns), address)
-
 	// Remove the port from the multicast allow policy.
 	if oc.multicastSupport && nsInfo.multicastEnabled {
 		if err := podDeleteAllowMulticastPolicy(ns, portInfo); err != nil {
@@ -91,6 +89,14 @@ func (oc *Controller) deletePodFromNamespace(ns string, portInfo *lpInfo) error
 	return nil
 }
 
+func createIPAddressSlice(ips []*net.IPNet) []net.IP {
+	ipAddrs := make([]net.IP, 0)
+	for _, ip := range ips {
+		ipAddrs = append(ipAddrs, ip.IP)
+	}
+	return ipAddrs
+}
+
 // Creates an explicit "allow" policy for multicast traffic within the
 // namespace if multicast is enabled. Otherwise, removes the "allow" policy.
 // Traffic will be dropped by the default multicast deny ACL.
@@ -107,28 +113,63 @@ func (oc *Controller) multicastUpdateNamespace(ns *kapi.Namespace, nsInfo *names
 	}
 
 	var err error
+	nsInfo.multicastEnabled = enabled
 	if enabled {
 		err = oc.createMulticastAllowPolicy(ns.Name, nsInfo)
 	} else {
-		err = deleteMulticastAllowPolicy(ns.Name)
+		err = deleteMulticastAllowPolicy(ns.Name, nsInfo)
 	}
 	if err != nil {
 		klog.Errorf(err.Error())
 		return
 	}
-
-	nsInfo.multicastEnabled = enabled
 }
 
 // Cleans up the multicast policy for this namespace if multicast was
 // previously allowed.
 func (oc *Controller) multicastDeleteNamespace(ns *kapi.Namespace, nsInfo *namespaceInfo) {
 	if nsInfo.multicastEnabled {
-		if err := deleteMulticastAllowPolicy(ns.Name); err != nil {
+		nsInfo.multicastEnabled = false
+		if err := deleteMulticastAllowPolicy(ns.Name, nsInfo); err != nil {
 			klog.Errorf(err.Error())
 		}
 	}
-	nsInfo.multicastEnabled = false
+}
+
+// updateNamepacePortGroup updates the port_group applied to the namespace. Multiple objects
+// that apply network configuration to all pods in a namespace will use the same port group.
+// This function ensures that the namespace wide port group will only be created once and
+// cleaned up when no object that relies on it exists.
+func (nsInfo *namespaceInfo) updateNamespacePortGroup(ns string) error {
+	if nsInfo.multicastEnabled {
+		if nsInfo.portGroupUUID != "" {
+			// Multicast is enabled and the port group exists so there is nothing to do.
+			return nil
+		}
+
+		// The port group should exist but doesn't so create it
+		portGroupUUID, err := createPortGroup(ns, hashedPortGroup(ns))
+		if err != nil {
+			return fmt.Errorf("failed to create port_group for %s (%v)", ns, err)
+		}
+		nsInfo.portGroupUUID = portGroupUUID
+	} else {
+		deletePortGroup(hashedPortGroup(ns))
+		nsInfo.portGroupUUID = ""
+	}
+	return nil
+}
+
+func parseRoutingExternalGWAnnotation(annotation string) ([]net.IP, error) {
+	var routingExternalGWs []net.IP
+	for _, v := range strings.Split(annotation, ",") {
+		parsedAnnotation := net.ParseIP(v)
+		if parsedAnnotation == nil {
+			return nil, fmt.Errorf("could not parse routing external gw annotation value %s", v)
+		}
+		routingExternalGWs = append(routingExternalGWs, parsedAnnotation)
+	}
+	return routingExternalGWs, nil
 }
 
 // AddNamespace creates corresponding addressset in ovn db
@@ -139,16 +180,20 @@ func (oc *Controller) AddNamespace(ns *kapi.Namespace) {
 
 	// Get all the pods in the namespace and append their IP to the
 	// address_set
+	var ips []net.IP
 	existingPods, err := oc.watchFactory.GetPods(ns.Name)
-	addresses := make([]string, 0, len(existingPods))
 	if err != nil {
 		klog.Errorf("Failed to get all the pods (%v)", err)
 	} else {
+		ips = make([]net.IP, 0, len(existingPods))
 		for _, pod := range existingPods {
 			if pod.Status.PodIP != "" && !pod.Spec.HostNetwork {
-				portName := podLogicalPortName(pod)
-				nsInfo.addressSet[pod.Status.PodIP] = portName
-				addresses = append(addresses, pod.Status.PodIP)
+				podIPs, err := util.GetAllPodIPs(pod)
+				if err != nil {
+					klog.Warningf(err.Error())
+					continue
+				}
+				ips = append(ips, podIPs...)
 			}
 		}
 	}
@@ -171,10 +216,22 @@ func (oc *Controller) AddNamespace(ns *kapi.Namespace) {
 			nsInfo.hybridOverlayVTEP = parsedAnnotation
 		}
 	}
+	annotation = ns.Annotations[routingExternalGWsAnnotation]
+	if annotation != "" {
+		nsInfo.routingExternalGWs, err = parseRoutingExternalGWAnnotation(annotation)
+		if err != nil {
+			klog.Errorf(err.Error())
+		}
+	}
+	nsInfo.addressSet, err = oc.addressSetFactory.NewAddressSet(ns.Name, ips)
+	if err != nil {
+		klog.Errorf(err.Error())
+	}
 
-	// Create an address_set for the namespace.  All the pods' IP address
-	// in the namespace will be added to the address_set
-	createAddressSet(ns.Name, hashedAddressSet(ns.Name), addresses)
+	// TODO(trozet) figure out if there is any possibility of detecting if a pod GW already exists, which
+	// is servicing this namespace. Right now that would mean searching through all pods, which is very inefficient.
+	// For now it is required that a pod serving as a gateway for a namespace is added AFTER the serving namespace is
+	// created
 
 	oc.multicastUpdateNamespace(ns, nsInfo)
 }
@@ -189,7 +246,59 @@ func (oc *Controller) updateNamespace(old, newer *kapi.Namespace) {
 	}
 	defer nsInfo.Unlock()
 
-	annotation := newer.Annotations[hotypes.HybridOverlayExternalGw]
+	var annotation, oldAnnotation string
+	annotation = newer.Annotations[routingExternalGWsAnnotation]
+	oldAnnotation = old.Annotations[routingExternalGWsAnnotation]
+	if annotation != oldAnnotation {
+		// if old gw annotation was empty, new one must not be empty, so we should remove any per pod SNAT
+		if oldAnnotation == "" {
+			if config.Gateway.DisableSNATMultipleGWs && (len(nsInfo.routingExternalGWs) != 0 || len(nsInfo.routingExternalPodGWs) != 0) {
+				existingPods, err := oc.watchFactory.GetPods(old.Name)
+				if err != nil {
+					klog.Errorf("Failed to get all the pods (%v)", err)
+				}
+				for _, pod := range existingPods {
+					logicalPort := podLogicalPortName(pod)
+					portInfo, err := oc.logicalPortCache.get(logicalPort)
+					if err != nil {
+						klog.Warningf("Unable to get port %s in cache for SNAT rule removal", logicalPort)
+					} else {
+						oc.deletePerPodGRSNAT(pod.Spec.NodeName, portInfo.ips)
+					}
+				}
+			}
+		} else {
+			oc.deleteGWRoutesForNamespace(nsInfo)
+		}
+		exGateways, err := parseRoutingExternalGWAnnotation(annotation)
+		if err != nil {
+			klog.Error(err.Error())
+		} else {
+			err = oc.addExternalGWsForNamespace(exGateways, nsInfo, old.Name)
+			if err != nil {
+				klog.Error(err.Error())
+			}
+		}
+		// if new annotation is empty, exgws were removed, may need to add SNAT per pod
+		// check if there are any pod gateways serving this namespace as well
+		if annotation == "" && len(nsInfo.routingExternalPodGWs) == 0 && config.Gateway.DisableSNATMultipleGWs {
+			existingPods, err := oc.watchFactory.GetPods(old.Name)
+			if err != nil {
+				klog.Errorf("Failed to get all the pods (%v)", err)
+			}
+			for _, pod := range existingPods {
+				podAnnotation, err := util.UnmarshalPodAnnotation(pod.Annotations)
+				if err != nil {
+					klog.Error(err.Error())
+				} else {
+					if err = oc.addPerPodGRSNAT(pod, podAnnotation.IPs); err != nil {
+						klog.Error(err.Error())
+					}
+				}
+			}
+		}
+	}
+	annotation = newer.Annotations[hotypes.HybridOverlayExternalGw]
 	if annotation != "" {
 		parsedAnnotation := net.ParseIP(annotation)
 		if parsedAnnotation == nil {
@@ -197,6 +306,8 @@ func (oc *Controller) updateNamespace(old, newer *kapi.Namespace) {
 		} else {
 			nsInfo.hybridOverlayExternalGW = parsedAnnotation
 		}
+	} else {
+		nsInfo.hybridOverlayExternalGW = nil
 	}
 	annotation = newer.Annotations[hotypes.HybridOverlayVTEP]
 	if annotation != "" {
@@ -206,6 +317,8 @@ func (oc *Controller) updateNamespace(old, newer *kapi.Namespace) {
 		} else {
 			nsInfo.hybridOverlayVTEP = parsedAnnotation
 		}
+	} else {
+		nsInfo.hybridOverlayVTEP = nil
 	}
 	oc.multicastUpdateNamespace(newer, nsInfo)
 }
@@ -218,14 +331,13 @@ func (oc *Controller) deleteNamespace(ns *kapi.Namespace) {
 		return
 	}
 	defer nsInfo.Unlock()
-	
+
 	klog.V(5).Infof("Deleting Namespace's NetworkPolicy entities")
 	for _, np := range nsInfo.networkPolicies {
 		delete(nsInfo.networkPolicies, np.name)
 		oc.destroyNamespacePolicy(np)
 	}
-	
-	deleteAddressSet(hashedAddressSet(ns.Name))
+	oc.deleteGWRoutesForNamespace(nsInfo)
 	oc.multicastDeleteNamespace(ns, nsInfo)
 }
 
@@ -281,9 +393,10 @@ func (oc *Controller) createNamespaceLocked(ns string) *namespaceInfo {
 	defer oc.namespacesMutex.Unlock()
 
 	nsInfo := &namespaceInfo{
-		addressSet:       make(map[string]string),
-		networkPolicies:  make(map[string]*namespacePolicy),
-		multicastEnabled: false,
+		networkPolicies:       make(map[string]*namespacePolicy),
+		podExternalRoutes:     make(map[string]map[string]string),
+		multicastEnabled:      false,
+		routingExternalPodGWs: make(map[string][]net.IP),
 	}
 	nsInfo.Lock()
 	oc.namespaces[ns] = nsInfo
@@ -311,6 +424,9 @@ func (oc *Controller) deleteNamespaceLocked(ns string) *namespaceInfo {
 		nsInfo.Unlock()
 		return nil
 	}
+	if err := nsInfo.addressSet.Destroy(); err != nil {
+		klog.Errorf(err.Error())
+	}
 	delete(oc.namespaces, ns)
 
 	return nsInfo
