@@ -10,23 +10,40 @@ import (
 	"sync"
 	"time"
 
+	goovn "github.com/ebay/go-ovn"
+	hocontroller "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/controller"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
+	egressipv1 "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressip/v1"
+	egressipapi "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressip/v1/apis/clientset/versioned"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/factory"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/kube"
-	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/ovn/allocator"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/ovn/ipallocator"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/ovn/subnetallocator"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
 
+	egressfirewall "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressfirewall/v1"
+	egressfirewallclientset "github.com/ovn-org/ovn-kubernetes/go-controller/pkg/crd/egressfirewall/v1/apis/clientset/versioned"
+
+	apiextension "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
+	utilnet "k8s.io/utils/net"
+
 	kapi "k8s.io/api/core/v1"
 	kapisnetworking "k8s.io/api/networking/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/kubernetes"
+	"k8s.io/client-go/kubernetes/scheme"
 	"k8s.io/client-go/tools/cache"
 	"k8s.io/client-go/tools/record"
+	ref "k8s.io/client-go/tools/reference"
 	"k8s.io/klog"
 )
 
+const (
+	egressfirewallCRD = "egressfirewalls.k8s.ovn.org"
+)
+
 // ServiceVIPKey is used for looking up service namespace information for a
 // particular load balancer
 type ServiceVIPKey struct {
@@ -51,8 +68,9 @@ type loadBalancerConf struct {
 type namespaceInfo struct {
 	sync.Mutex
 
-	// map from pod IP address to logical port name for all pods
-	addressSet map[string]string
+	// addressSet is an address set object that holds the IP addresses
+	// of all pods in the namespace.
+	addressSet AddressSet
 
 	// map from NetworkPolicy name to namespacePolicy. You must hold the
 	// namespaceInfo's mutex to add/delete/lookup policies, but must hold the
@@ -60,21 +78,54 @@ type namespaceInfo struct {
 	// the policy itself.
 	networkPolicies map[string]*namespacePolicy
 
+	//defines the namespaces egressFirewallPolicy
+	egressFirewallPolicy *egressFirewall
+
 	hybridOverlayExternalGW net.IP
 	hybridOverlayVTEP       net.IP
 
+	// routingExternalGWs is a slice of net.IP containing the values parsed from
+	// annotation k8s.ovn.org/routing-external-gws
+	routingExternalGWs []net.IP
+	// podExternalRoutes is a cache keeping the LR routes added to the GRs when
+	// the k8s.ovn.org/routing-external-gws annotation is used. The first map key
+	// is the podIP, the second the GW and the third the GR
+	podExternalRoutes map[string]map[string]string
+
+	// routingExternalPodGWs contains a map of all pods serving as exgws as well as their
+	// exgw IPs
+	routingExternalPodGWs map[string][]net.IP
+
+	// The UUID of the namespace-wide port group that contains all the pods in the namespace.
+	portGroupUUID string
+
 	multicastEnabled bool
 }
 
+// eNode is a cache helper used for egress IP assignment
+type eNode struct {
+	v4Subnet    *net.IPNet
+	v6Subnet    *net.IPNet
+	allocations map[string]bool
+	tainted     bool
+	name        string
+}
+
 // Controller structure is the object which holds the controls for starting
 // and reacting upon the watched resources (e.g. pods, endpoints)
 type Controller struct {
-	kube         kube.Interface
-	watchFactory *factory.WatchFactory
-	stopChan     <-chan struct{}
+	kube                  kube.Interface
+	watchFactory          *factory.WatchFactory
+	egressFirewallHandler *factory.Handler
+	stopChan              <-chan struct{}
 
-	masterSubnetAllocator *allocator.SubnetAllocator
-	joinSubnetAllocator   *allocator.SubnetAllocator
+	// FIXME DUAL-STACK -  Make IP Allocators more dual-stack friendly
+	masterSubnetAllocator     *subnetallocator.SubnetAllocator
+	joinSubnetAllocator       *subnetallocator.SubnetAllocator
+	nodeLocalNatIPv4Allocator *ipallocator.Range
+	nodeLocalNatIPv6Allocator *ipallocator.Range
+
+	hoMaster *hocontroller.MasterController
 
 	TCPLoadBalancerUUID  string
 	UDPLoadBalancerUUID  string
@@ -85,13 +136,8 @@ type Controller struct {
 	// cluster's east-west traffic.
 	loadbalancerClusterCache map[kapi.Protocol]string
 
-	// For TCP and UDP type traffice, cache OVN load balancer that exists on the
-	// default gateway
-	loadbalancerGWCache map[kapi.Protocol]string
-	defGatewayRouter    string
-
 	// A cache of all logical switches seen by the watcher and their subnets
-	logicalSwitchCache map[string][]*net.IPNet
+	lsManager *logicalSwitchManager
 
 	// A cache of all logical ports known to the controller
 	logicalPortCache *portCache
@@ -103,6 +149,9 @@ type Controller struct {
 	namespaces      map[string]*namespaceInfo
 	namespacesMutex sync.Mutex
 
+	// An address set factory that creates address sets
+	addressSetFactory AddressSetFactory
+
 	// Port group for ingress deny rule
 	portGroupIngressDeny string
 
@@ -120,12 +169,34 @@ type Controller struct {
 	// A mutex for lspIngressDenyCache and lspEgressDenyCache
 	lspMutex *sync.Mutex
 
-	// A mutex for logicalSwitchCache which holds logicalSwitch information
-	lsMutex *sync.Mutex
-
 	// Supports multicast?
 	multicastSupport bool
 
+	// Interface used for programming OVN for egress IP, based on the mode it's running in.
+	modeEgressIP modeEgressIP
+
+	// Sync used for retrying EgressIP objects which were created before any node existed.
+	egressAssignmentRetry sync.Map
+
+	// Mutex used for syncing the egressIP namespace handlers
+	egressIPNamespaceHandlerMutex *sync.Mutex
+
+	// Cache used for keeping track of EgressIP namespace handlers
+	egressIPNamespaceHandlerCache map[string]factory.Handler
+
+	// Mutex used for syncing the egressIP pod handlers
+	egressIPPodHandlerMutex *sync.Mutex
+
+	// Cache used for keeping track of EgressIP pod handlers
+	egressIPPodHandlerCache map[string]factory.Handler
+
+	// A cache used for egress IP assignments containing data for all cluster nodes
+	// used for egress IP assignments
+	eIPAllocator map[string]*eNode
+
+	// A mutex for eIPAllocator
+	eIPAllocatorMutex *sync.Mutex
+
 	// Map of load balancers to service namespace
 	serviceVIPToName map[ServiceVIPKey]types.NamespacedName
 
@@ -138,6 +209,12 @@ type Controller struct {
 
 	// event recorder used to post events to k8s
 	recorder record.EventRecorder
+
+	// go-ovn northbound client interface
+	ovnNBClient goovn.Client
+
+	// go-ovn southbound client interface
+	ovnSBClient goovn.Client
 }
 
 const (
@@ -151,55 +228,108 @@ const (
 	SCTP = "SCTP"
 )
 
+func GetIPFullMask(ip string) string {
+	const (
+		// IPv4FullMask is the maximum prefix mask for an IPv4 address
+		IPv4FullMask = "/32"
+		// IPv6FullMask is the maxiumum prefix mask for an IPv6 address
+		IPv6FullMask = "/128"
+	)
+
+	if utilnet.IsIPv6(net.ParseIP(ip)) {
+		return IPv6FullMask
+	}
+	return IPv4FullMask
+}
+
 // NewOvnController creates a new OVN controller for creating logical network
 // infrastructure and policy
-func NewOvnController(kubeClient kubernetes.Interface, wf *factory.WatchFactory, stopChan <-chan struct{}) *Controller {
+func NewOvnController(kubeClient kubernetes.Interface, egressIPClient egressipapi.Interface, egressFirewallClient egressfirewallclientset.Interface, wf *factory.WatchFactory,
+	stopChan <-chan struct{}, addressSetFactory AddressSetFactory, ovnNBClient goovn.Client, ovnSBClient goovn.Client, recorder record.EventRecorder) *Controller {
+
+	if addressSetFactory == nil {
+		addressSetFactory = NewOvnAddressSetFactory()
+	}
+	modeEgressIP := newModeEgressIP()
 	return &Controller{
-		kube:                     &kube.Kube{KClient: kubeClient},
-		watchFactory:             wf,
-		stopChan:                 stopChan,
-		masterSubnetAllocator:    allocator.NewSubnetAllocator(),
-		logicalSwitchCache:       make(map[string][]*net.IPNet),
-		joinSubnetAllocator:      allocator.NewSubnetAllocator(),
-		logicalPortCache:         newPortCache(stopChan),
-		namespaces:               make(map[string]*namespaceInfo),
-		namespacesMutex:          sync.Mutex{},
-		lspIngressDenyCache:      make(map[string]int),
-		lspEgressDenyCache:       make(map[string]int),
-		lspMutex:                 &sync.Mutex{},
-		lsMutex:                  &sync.Mutex{},
-		loadbalancerClusterCache: make(map[kapi.Protocol]string),
-		loadbalancerGWCache:      make(map[kapi.Protocol]string),
-		multicastSupport:         config.EnableMulticast,
-		serviceVIPToName:         make(map[ServiceVIPKey]types.NamespacedName),
-		serviceVIPToNameLock:     sync.Mutex{},
-		serviceLBMap:             make(map[string]map[string]*loadBalancerConf),
-		serviceLBLock:            sync.Mutex{},
-		recorder:                 util.EventRecorder(kubeClient),
+		kube: &kube.Kube{
+			KClient:              kubeClient,
+			EIPClient:            egressIPClient,
+			EgressFirewallClient: egressFirewallClient,
+		},
+		watchFactory:                  wf,
+		stopChan:                      stopChan,
+		masterSubnetAllocator:         subnetallocator.NewSubnetAllocator(),
+		nodeLocalNatIPv4Allocator:     &ipallocator.Range{},
+		nodeLocalNatIPv6Allocator:     &ipallocator.Range{},
+		lsManager:                     newLogicalSwitchManager(),
+		joinSubnetAllocator:           subnetallocator.NewSubnetAllocator(),
+		logicalPortCache:              newPortCache(stopChan),
+		namespaces:                    make(map[string]*namespaceInfo),
+		namespacesMutex:               sync.Mutex{},
+		addressSetFactory:             addressSetFactory,
+		lspIngressDenyCache:           make(map[string]int),
+		lspEgressDenyCache:            make(map[string]int),
+		lspMutex:                      &sync.Mutex{},
+		modeEgressIP:                  modeEgressIP,
+		egressIPNamespaceHandlerMutex: &sync.Mutex{},
+		egressIPNamespaceHandlerCache: make(map[string]factory.Handler),
+		egressIPPodHandlerMutex:       &sync.Mutex{},
+		egressIPPodHandlerCache:       make(map[string]factory.Handler),
+		eIPAllocatorMutex:             &sync.Mutex{},
+		eIPAllocator:                  make(map[string]*eNode),
+		loadbalancerClusterCache:      make(map[kapi.Protocol]string),
+		multicastSupport:              config.EnableMulticast,
+		serviceVIPToName:              make(map[ServiceVIPKey]types.NamespacedName),
+		serviceVIPToNameLock:          sync.Mutex{},
+		serviceLBMap:                  make(map[string]map[string]*loadBalancerConf),
+		serviceLBLock:                 sync.Mutex{},
+		recorder:                      recorder,
+		ovnNBClient:                   ovnNBClient,
+		ovnSBClient:                   ovnSBClient,
 	}
 }
 
 // Run starts the actual watching.
-func (oc *Controller) Run() error {
+func (oc *Controller) Run(wg *sync.WaitGroup) error {
 	oc.syncPeriodic()
-	// WatchNodes must be started first so that its initial Add will
-	// create all node logical switches, which other watches may depend on.
+	klog.Infof("Starting all the Watchers...")
+	start := time.Now()
+
+	// WatchNamespaces() should be started first because it has no other
+	// dependencies, and WatchNodes() depends on it
+	oc.WatchNamespaces()
+
+	// WatchNodes must be started next because it creates the node switch
+	// which most other watches depend on.
 	// https://github.com/ovn-org/ovn-kubernetes/pull/859
-	if err := oc.WatchNodes(); err != nil {
-		return err
+	oc.WatchNodes()
+
+	oc.WatchPods()
+	oc.WatchServices()
+	oc.WatchEndpoints()
+	oc.WatchNetworkPolicy()
+	oc.WatchCRD()
+
+	if config.OVNKubernetesFeature.EnableEgressIP {
+		oc.WatchEgressNodes()
+		oc.WatchEgressIP()
 	}
 
-	for _, f := range []func() error{oc.WatchPods, oc.WatchServices, oc.WatchEndpoints,
-		oc.WatchNamespaces, oc.WatchNetworkPolicy} {
-		if err := f(); err != nil {
-			return err
-		}
-	}
+	klog.Infof("Completing all the Watchers took %v", time.Since(start))
 
 	if config.Kubernetes.OVNEmptyLbEvents {
 		go oc.ovnControllerEventChecker()
 	}
 
+	if oc.hoMaster != nil {
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			oc.hoMaster.Run(oc.stopChan)
+		}()
+	}
+
 	return nil
 }
 
@@ -214,6 +344,10 @@ type emptyLBBackendEvent struct {
 	uuid     string
 }
 
+func newModeEgressIP() modeEgressIP {
+	return &egressIPMode{}
+}
+
 func extractEmptyLBBackendsEvents(out []byte) ([]emptyLBBackendEvent, error) {
 	events := make([]emptyLBBackendEvent, 0, 4)
 
@@ -381,19 +515,36 @@ func podScheduled(pod *kapi.Pod) bool {
 	return pod.Spec.NodeName != ""
 }
 
+func (oc *Controller) recordPodEvent(addErr error, pod *kapi.Pod) {
+	podRef, err := ref.GetReference(scheme.Scheme, pod)
+	if err != nil {
+		klog.Errorf("Couldn't get a reference to pod %s/%s to post an event: '%v'",
+			pod.Namespace, pod.Name, err)
+	} else {
+		klog.V(5).Infof("Posting a %s event for Pod %s/%s", kapi.EventTypeWarning, pod.Namespace, pod.Name)
+		oc.recorder.Eventf(podRef, kapi.EventTypeWarning, "ErrorAddingLogicalPort", addErr.Error())
+	}
+}
+
 // WatchPods starts the watching of Pod resource and calls back the appropriate handler logic
-func (oc *Controller) WatchPods() error {
+func (oc *Controller) WatchPods() {
 	var retryPods sync.Map
-	_, err := oc.watchFactory.AddPodHandler(cache.ResourceEventHandlerFuncs{
+
+	start := time.Now()
+	oc.watchFactory.AddPodHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			pod := obj.(*kapi.Pod)
 			if !podWantsNetwork(pod) {
+				// host network pod is able to serve as external gw for other pods
+				if err := oc.addPodExternalGW(pod); err != nil {
+					klog.Errorf(err.Error())
+				}
 				return
 			}
-
 			if podScheduled(pod) {
 				if err := oc.addLogicalPort(pod); err != nil {
 					klog.Errorf(err.Error())
+					oc.recordPodEvent(err, pod)
 					retryPods.Store(pod.UID, true)
 				}
 			} else {
@@ -402,33 +553,49 @@ func (oc *Controller) WatchPods() error {
 			}
 		},
 		UpdateFunc: func(old, newer interface{}) {
+			oldPod := old.(*kapi.Pod)
 			pod := newer.(*kapi.Pod)
-			if !podWantsNetwork(pod) {
-				return
-			}
 
 			_, retry := retryPods.Load(pod.UID)
-			if podScheduled(pod) && retry {
+			if podScheduled(pod) && retry && podWantsNetwork(pod) {
 				if err := oc.addLogicalPort(pod); err != nil {
 					klog.Errorf(err.Error())
+					oc.recordPodEvent(err, pod)
 				} else {
 					retryPods.Delete(pod.UID)
 				}
+			} else {
+				// No matter if a pod is ovn networked, or host networked, we still need to check for exgw
+				// annotations. If the pod is ovn networked and is in update reschedule, addLogicalPort will take
+				// care of updating the exgw updates
+				if oldPod.Annotations[routingNamespaceAnnotation] != pod.Annotations[routingNamespaceAnnotation] ||
+					oldPod.Annotations[routingNetworkAnnotation] != pod.Annotations[routingNetworkAnnotation] {
+					oc.deletePodExternalGW(oldPod)
+					if err := oc.addPodExternalGW(pod); err != nil {
+						klog.Errorf(err.Error())
+					}
+				}
 			}
 		},
 		DeleteFunc: func(obj interface{}) {
 			pod := obj.(*kapi.Pod)
+			if !podWantsNetwork(pod) {
+				oc.deletePodExternalGW(pod)
+				return
+			}
+			// deleteLogicalPort will take care of removing exgw for ovn networked pods
 			oc.deleteLogicalPort(pod)
 			retryPods.Delete(pod.UID)
 		},
 	}, oc.syncPods)
-	return err
+	klog.Infof("Bootstrapping existing pods and cleaning stale pods took %v", time.Since(start))
 }
 
 // WatchServices starts the watching of Service resource and calls back the
 // appropriate handler logic
-func (oc *Controller) WatchServices() error {
-	_, err := oc.watchFactory.AddServiceHandler(cache.ResourceEventHandlerFuncs{
+func (oc *Controller) WatchServices() {
+	start := time.Now()
+	oc.watchFactory.AddServiceHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			service := obj.(*kapi.Service)
 			err := oc.createService(service)
@@ -449,12 +616,13 @@ func (oc *Controller) WatchServices() error {
 			oc.deleteService(service)
 		},
 	}, oc.syncServices)
-	return err
+	klog.Infof("Bootstrapping existing services and cleaning stale services took %v", time.Since(start))
 }
 
 // WatchEndpoints starts the watching of Endpoint resource and calls back the appropriate handler logic
-func (oc *Controller) WatchEndpoints() error {
-	_, err := oc.watchFactory.AddEndpointsHandler(cache.ResourceEventHandlerFuncs{
+func (oc *Controller) WatchEndpoints() {
+	start := time.Now()
+	oc.watchFactory.AddEndpointsHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			ep := obj.(*kapi.Endpoints)
 			err := oc.AddEndpoints(ep)
@@ -488,13 +656,14 @@ func (oc *Controller) WatchEndpoints() error {
 			}
 		},
 	}, nil)
-	return err
+	klog.Infof("Bootstrapping existing endpoints and cleaning stale endpoints took %v", time.Since(start))
 }
 
 // WatchNetworkPolicy starts the watching of network policy resource and calls
 // back the appropriate handler logic
-func (oc *Controller) WatchNetworkPolicy() error {
-	_, err := oc.watchFactory.AddPolicyHandler(cache.ResourceEventHandlerFuncs{
+func (oc *Controller) WatchNetworkPolicy() {
+	start := time.Now()
+	oc.watchFactory.AddPolicyHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			policy := obj.(*kapisnetworking.NetworkPolicy)
 			oc.addNetworkPolicy(policy)
@@ -512,13 +681,184 @@ func (oc *Controller) WatchNetworkPolicy() error {
 			oc.deleteNetworkPolicy(policy)
 		},
 	}, oc.syncNetworkPolicies)
-	return err
+	klog.Infof("Bootstrapping existing policies and cleaning stale policies took %v", time.Since(start))
+}
+
+// WatchCRD starts the watching of the CRD resource and calls back to the
+// appropriate handler logic
+func (oc *Controller) WatchCRD() {
+	oc.watchFactory.AddCRDHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc: func(obj interface{}) {
+			crd := obj.(*apiextension.CustomResourceDefinition)
+			klog.Infof("Adding CRD %s to cluster", crd.Name)
+			if crd.Name == egressfirewallCRD {
+				err := oc.watchFactory.InitializeEgressFirewallWatchFactory()
+				if err != nil {
+					klog.Errorf("Error Creating EgressFirewallWatchFactory: %v", err)
+					return
+				}
+				oc.egressFirewallHandler = oc.WatchEgressFirewall()
+
+			}
+		},
+		UpdateFunc: func(old, newer interface{}) {
+		},
+		DeleteFunc: func(obj interface{}) {
+			crd := obj.(*apiextension.CustomResourceDefinition)
+			klog.Infof("Deleting CRD %s from cluster", crd.Name)
+			if crd.Name == egressfirewallCRD {
+				oc.watchFactory.RemoveEgressFirewallHandler(oc.egressFirewallHandler)
+				oc.egressFirewallHandler = nil
+				oc.watchFactory.ShutdownEgressFirewallWatchFactory()
+			}
+		},
+	}, nil)
+}
+
+// WatchEgressFirewall starts the watching of egressfirewall resource and calls
+// back the appropriate handler logic
+func (oc *Controller) WatchEgressFirewall() *factory.Handler {
+	return oc.watchFactory.AddEgressFirewallHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc: func(obj interface{}) {
+			egressFirewall := obj.(*egressfirewall.EgressFirewall).DeepCopy()
+			errList := oc.addEgressFirewall(egressFirewall)
+			for _, err := range errList {
+				klog.Error(err)
+			}
+			if len(errList) == 0 {
+				egressFirewall.Status.Status = egressFirewallAppliedCorrectly
+			} else {
+				egressFirewall.Status.Status = egressFirewallAddError
+			}
+			err := oc.updateEgressFirewallWithRetry(egressFirewall)
+			if err != nil {
+				klog.Error(err)
+			}
+		},
+		UpdateFunc: func(old, newer interface{}) {
+			newEgressFirewall := newer.(*egressfirewall.EgressFirewall).DeepCopy()
+			oldEgressFirewall := old.(*egressfirewall.EgressFirewall)
+			if !reflect.DeepEqual(oldEgressFirewall.Spec, newEgressFirewall.Spec) {
+				errList := oc.updateEgressFirewall(oldEgressFirewall, newEgressFirewall)
+				if len(errList) > 0 {
+					newEgressFirewall.Status.Status = egressFirewallUpdateError
+					for _, err := range errList {
+						klog.Error(err)
+					}
+				} else {
+					newEgressFirewall.Status.Status = egressFirewallAppliedCorrectly
+				}
+				err := oc.updateEgressFirewallWithRetry(newEgressFirewall)
+				if err != nil {
+					klog.Error(err)
+				}
+			}
+		},
+		DeleteFunc: func(obj interface{}) {
+			egressFirewall := obj.(*egressfirewall.EgressFirewall)
+			errList := oc.deleteEgressFirewall(egressFirewall)
+			for _, err := range errList {
+				klog.Error(err)
+			}
+		},
+	}, nil)
+}
+
+// WatchEgressNodes starts the watching of egress assignable nodes and calls
+// back the appropriate handler logic.
+func (oc *Controller) WatchEgressNodes() {
+	nodeEgressLabel := util.GetNodeEgressLabel()
+	oc.watchFactory.AddNodeHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc: func(obj interface{}) {
+			node := obj.(*kapi.Node)
+			if err := oc.addNodeForEgress(node); err != nil {
+				klog.Error(err)
+			}
+			nodeLabels := node.GetLabels()
+			if _, hasEgressLabel := nodeLabels[nodeEgressLabel]; hasEgressLabel {
+				if err := oc.addEgressNode(node); err != nil {
+					klog.Error(err)
+				}
+			}
+		},
+		UpdateFunc: func(old, new interface{}) {
+			oldNode := old.(*kapi.Node)
+			newNode := new.(*kapi.Node)
+			oldLabels := oldNode.GetLabels()
+			newLabels := newNode.GetLabels()
+			_, oldHadEgressLabel := oldLabels[nodeEgressLabel]
+			_, newHasEgressLabel := newLabels[nodeEgressLabel]
+			if !oldHadEgressLabel && newHasEgressLabel {
+				if err := oc.addEgressNode(newNode); err != nil {
+					klog.Error(err)
+				}
+			}
+			if oldHadEgressLabel && !newHasEgressLabel {
+				if err := oc.deleteEgressNode(oldNode); err != nil {
+					klog.Error(err)
+				}
+			}
+		},
+		DeleteFunc: func(obj interface{}) {
+			node := obj.(*kapi.Node)
+			if err := oc.deleteNodeForEgress(node); err != nil {
+				klog.Error(err)
+			}
+			nodeLabels := node.GetLabels()
+			if _, hasEgressLabel := nodeLabels[nodeEgressLabel]; hasEgressLabel {
+				if err := oc.deleteEgressNode(node); err != nil {
+					klog.Error(err)
+				}
+			}
+		},
+	}, oc.initClusterEgressPolicies)
+}
+
+// WatchEgressIP starts the watching of egressip resource and calls
+// back the appropriate handler logic.
+func (oc *Controller) WatchEgressIP() {
+	oc.watchFactory.AddEgressIPHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc: func(obj interface{}) {
+			eIP := obj.(*egressipv1.EgressIP).DeepCopy()
+			if err := oc.addEgressIP(eIP); err != nil {
+				klog.Error(err)
+			}
+			if err := oc.updateEgressIPWithRetry(eIP); err != nil {
+				klog.Error(err)
+			}
+		},
+		UpdateFunc: func(old, new interface{}) {
+			oldEIP := old.(*egressipv1.EgressIP)
+			newEIP := new.(*egressipv1.EgressIP).DeepCopy()
+			if !reflect.DeepEqual(oldEIP.Spec, newEIP.Spec) {
+				if err := oc.deleteEgressIP(oldEIP); err != nil {
+					klog.Error(err)
+				}
+				newEIP.Status = egressipv1.EgressIPStatus{
+					Items: []egressipv1.EgressIPStatusItem{},
+				}
+				if err := oc.addEgressIP(newEIP); err != nil {
+					klog.Error(err)
+				}
+				if err := oc.updateEgressIPWithRetry(newEIP); err != nil {
+					klog.Error(err)
+				}
+			}
+		},
+		DeleteFunc: func(obj interface{}) {
+			eIP := obj.(*egressipv1.EgressIP)
+			if err := oc.deleteEgressIP(eIP); err != nil {
+				klog.Error(err)
+			}
+		},
+	}, oc.syncEgressIPs)
 }
 
 // WatchNamespaces starts the watching of namespace resource and calls
 // back the appropriate handler logic
-func (oc *Controller) WatchNamespaces() error {
-	_, err := oc.watchFactory.AddNamespaceHandler(cache.ResourceEventHandlerFuncs{
+func (oc *Controller) WatchNamespaces() {
+	start := time.Now()
+	oc.watchFactory.AddNamespaceHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			ns := obj.(*kapi.Namespace)
 			oc.AddNamespace(ns)
@@ -532,7 +872,7 @@ func (oc *Controller) WatchNamespaces() error {
 			oc.deleteNamespace(ns)
 		},
 	}, oc.syncNamespaces)
-	return err
+	klog.Infof("Bootstrapping existing namespaces and cleaning stale namespaces took %v", time.Since(start))
 }
 
 func (oc *Controller) syncNodeGateway(node *kapi.Node, hostSubnets []*net.IPNet) error {
@@ -545,7 +885,7 @@ func (oc *Controller) syncNodeGateway(node *kapi.Node, hostSubnets []*net.IPNet)
 		hostSubnets, _ = util.ParseNodeHostSubnetAnnotation(node)
 	}
 	if l3GatewayConfig.Mode == config.GatewayModeDisabled {
-		if err := gatewayCleanup(node.Name, hostSubnets); err != nil {
+		if err := gatewayCleanup(node.Name); err != nil {
 			return fmt.Errorf("error cleaning up gateway for node %s: %v", node.Name, err)
 		}
 	} else if hostSubnets != nil {
@@ -558,35 +898,45 @@ func (oc *Controller) syncNodeGateway(node *kapi.Node, hostSubnets []*net.IPNet)
 
 // WatchNodes starts the watching of node resource and calls
 // back the appropriate handler logic
-func (oc *Controller) WatchNodes() error {
+func (oc *Controller) WatchNodes() {
 	var gatewaysFailed sync.Map
 	var mgmtPortFailed sync.Map
-	_, err := oc.watchFactory.AddNodeHandler(cache.ResourceEventHandlerFuncs{
+	var addNodeFailed sync.Map
+
+	start := time.Now()
+	oc.watchFactory.AddNodeHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(obj interface{}) {
 			node := obj.(*kapi.Node)
 			if noHostSubnet := noHostSubnet(node); noHostSubnet {
-				oc.lsMutex.Lock()
-				defer oc.lsMutex.Unlock()
-				//setting the value to nil in the cache means it was not assigned a hostSubnet by ovn-kube
-				oc.logicalSwitchCache[node.Name] = nil
+				err := oc.lsManager.AddNoHostSubnetNode(node.Name)
+				if err != nil {
+					klog.Errorf("Error creating logical switch cache for node %s: %v", node.Name, err)
+				}
 				return
 			}
 
 			klog.V(5).Infof("Added event for Node %q", node.Name)
 			hostSubnets, err := oc.addNode(node)
 			if err != nil {
-				klog.Errorf("error creating subnet for node %s: %v", node.Name, err)
+				klog.Errorf("NodeAdd: error creating subnet for node %s: %v", node.Name, err)
+				addNodeFailed.Store(node.Name, true)
+				mgmtPortFailed.Store(node.Name, true)
+				gatewaysFailed.Store(node.Name, true)
 				return
 			}
 
 			err = oc.syncNodeManagementPort(node, hostSubnets)
 			if err != nil {
-				klog.Warningf("error creating management port for node %s: %v", node.Name, err)
+				if !util.IsAnnotationNotSetError(err) {
+					klog.Warningf("Error creating management port for node %s: %v", node.Name, err)
+				}
 				mgmtPortFailed.Store(node.Name, true)
 			}
 
 			if err := oc.syncNodeGateway(node, hostSubnets); err != nil {
-				klog.Warningf(err.Error())
+				if !util.IsAnnotationNotSetError(err) {
+					klog.Warningf(err.Error())
+				}
 				gatewaysFailed.Store(node.Name, true)
 			}
 		},
@@ -603,13 +953,24 @@ func (oc *Controller) WatchNodes() error {
 				return
 			}
 
-			klog.V(5).Infof("Updated event for Node %q", node.Name)
-
-			_, failed := mgmtPortFailed.Load(node.Name)
-			if failed || macAddressChanged(oldNode, node) {
-				err := oc.syncNodeManagementPort(node, nil)
+			var hostSubnets []*net.IPNet
+			_, failed := addNodeFailed.Load(node.Name)
+			if failed {
+				hostSubnets, err = oc.addNode(node)
 				if err != nil {
-					klog.Errorf("error updating management port for node %s: %v", node.Name, err)
+					klog.Errorf("NodeUpdate: error creating subnet for node %s: %v", node.Name, err)
+					return
+				}
+				addNodeFailed.Delete(node.Name)
+			}
+
+			_, failed = mgmtPortFailed.Load(node.Name)
+			if failed || macAddressChanged(oldNode, node) {
+				err := oc.syncNodeManagementPort(node, hostSubnets)
+				if err != nil {
+					if !util.IsAnnotationNotSetError(err) {
+						klog.Errorf("Error updating management port for node %s: %v", node.Name, err)
+					}
 					mgmtPortFailed.Store(node.Name, true)
 				} else {
 					mgmtPortFailed.Delete(node.Name)
@@ -622,7 +983,9 @@ func (oc *Controller) WatchNodes() error {
 			if failed || gatewayChanged(oldNode, node) {
 				err := oc.syncNodeGateway(node, nil)
 				if err != nil {
-					klog.Errorf(err.Error())
+					if !util.IsAnnotationNotSetError(err) {
+						klog.Errorf(err.Error())
+					}
 					gatewaysFailed.Store(node.Name, true)
 				} else {
 					gatewaysFailed.Delete(node.Name)
@@ -636,26 +999,18 @@ func (oc *Controller) WatchNodes() error {
 
 			nodeSubnets, _ := util.ParseNodeHostSubnetAnnotation(node)
 			joinSubnets, _ := util.ParseNodeJoinSubnetAnnotation(node)
-			err := oc.deleteNode(node.Name, nodeSubnets, joinSubnets)
+			dnatSnatIPs, _ := util.ParseNodeLocalNatIPAnnotation(node)
+			err := oc.deleteNode(node.Name, nodeSubnets, joinSubnets, dnatSnatIPs)
 			if err != nil {
 				klog.Error(err)
 			}
-			oc.lsMutex.Lock()
-			delete(oc.logicalSwitchCache, node.Name)
-			oc.lsMutex.Unlock()
+			oc.lsManager.DeleteNode(node.Name)
+			addNodeFailed.Delete(node.Name)
 			mgmtPortFailed.Delete(node.Name)
 			gatewaysFailed.Delete(node.Name)
-			// If this node was serving the external IP load balancer for services, migrate to a new node
-			if oc.defGatewayRouter == gwRouterPrefix+node.Name {
-				delete(oc.loadbalancerGWCache, kapi.ProtocolTCP)
-				delete(oc.loadbalancerGWCache, kapi.ProtocolUDP)
-				delete(oc.loadbalancerGWCache, kapi.ProtocolSCTP)
-				oc.defGatewayRouter = ""
-				oc.updateExternalIPsLB()
-			}
 		},
 	}, oc.syncNodes)
-	return err
+	klog.Infof("Bootstrapping existing nodes and cleaning stale nodes took %v", time.Since(start))
 }
 
 // AddServiceVIPToName associates a k8s service name with a load balancer VIP
