@@ -1,8 +1,16 @@
 package ovn
 
 import (
+	"context"
+	"encoding/json"
 	"fmt"
 	"net"
+	"strings"
+	"time"
+
+	"k8s.io/client-go/kubernetes/fake"
+
+	hotypes "github.com/ovn-org/ovn-kubernetes/go-controller/hybrid-overlay/pkg/types"
 
 	"github.com/urfave/cli/v2"
 
@@ -18,20 +26,34 @@ import (
 	. "github.com/onsi/gomega"
 )
 
-func newPodMeta(namespace, name string) metav1.ObjectMeta {
+func getPodAnnotations(fakeClient *fake.Clientset, namespace, name string) string {
+	pod, err := fakeClient.CoreV1().Pods(namespace).Get(context.TODO(), name, metav1.GetOptions{})
+	Expect(err).NotTo(HaveOccurred())
+	return pod.Annotations[util.OvnPodAnnotationName]
+}
+
+func newPodMeta(namespace, name string, additionalLabels map[string]string) metav1.ObjectMeta {
+	labels := map[string]string{
+		"name": name,
+	}
+	for k, v := range additionalLabels {
+		labels[k] = v
+	}
 	return metav1.ObjectMeta{
 		Name:      name,
 		UID:       types.UID(name),
 		Namespace: namespace,
-		Labels: map[string]string{
-			"name": name,
-		},
+		Labels:    labels,
 	}
 }
 
-func newPod(namespace, name, node, podIP string) *v1.Pod {
+func newPodWithLabels(namespace, name, node, podIP string, additionalLabels map[string]string) *v1.Pod {
+	podIPs := []v1.PodIP{}
+	if podIP != "" {
+		podIPs = append(podIPs, v1.PodIP{IP: podIP})
+	}
 	return &v1.Pod{
-		ObjectMeta: newPodMeta(namespace, name),
+		ObjectMeta: newPodMeta(namespace, name, additionalLabels),
 		Spec: v1.PodSpec{
 			Containers: []v1.Container{
 				{
@@ -42,8 +64,33 @@ func newPod(namespace, name, node, podIP string) *v1.Pod {
 			NodeName: node,
 		},
 		Status: v1.PodStatus{
-			Phase: v1.PodRunning,
-			PodIP: podIP,
+			Phase:  v1.PodRunning,
+			PodIP:  podIP,
+			PodIPs: podIPs,
+		},
+	}
+}
+
+func newPod(namespace, name, node, podIP string) *v1.Pod {
+	podIPs := []v1.PodIP{}
+	if podIP != "" {
+		podIPs = append(podIPs, v1.PodIP{IP: podIP})
+	}
+	return &v1.Pod{
+		ObjectMeta: newPodMeta(namespace, name, nil),
+		Spec: v1.PodSpec{
+			Containers: []v1.Container{
+				{
+					Name:  "containerName",
+					Image: "containerImage",
+				},
+			},
+			NodeName: node,
+		},
+		Status: v1.PodStatus{
+			Phase:  v1.PodRunning,
+			PodIP:  podIP,
+			PodIPs: podIPs,
 		},
 	}
 }
@@ -84,92 +131,57 @@ func (p pod) baseCmds(fexec *ovntest.FakeExec) {
 
 func (p pod) populateLogicalSwitchCache(fakeOvn *FakeOVN) {
 	Expect(p.nodeName).NotTo(Equal(""))
-	fakeOvn.controller.lsMutex.Lock()
-	defer fakeOvn.controller.lsMutex.Unlock()
-	fakeOvn.controller.logicalSwitchCache[p.nodeName] = []*net.IPNet{ovntest.MustParseIPNet(p.nodeSubnet)}
+	fakeOvn.controller.lsManager.AddNode(p.nodeName, []*net.IPNet{ovntest.MustParseIPNet(p.nodeSubnet)})
 }
 
 func (p pod) addCmds(fexec *ovntest.FakeExec, fail bool) {
 	// pod setup
-	fexec.AddFakeCmdsNoOutputNoError([]string{
-		"ovn-nbctl --timeout=15 --may-exist lsp-add " + p.nodeName + " " + p.portName + " -- lsp-set-addresses " + p.portName + " dynamic -- set logical_switch_port " + p.portName + " external-ids:namespace=" + p.namespace + " external-ids:pod=true",
-	})
-	fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-		Cmd:    "ovn-nbctl --timeout=15 get logical_switch_port " + p.portName + " dynamic_addresses addresses",
-		Output: `"` + p.podMAC + " " + p.podIP + `"` + "\n" + "[]",
-	})
-	fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-		Cmd:    "ovn-nbctl --timeout=15 get logical_switch_port " + p.portName + " _uuid",
-		Output: fakeUUID + "\n",
-	})
-	if fail {
-		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-			Cmd: "ovn-nbctl --timeout=15 lsp-set-port-security " + p.portName + " " + p.podMAC + " " + p.podIP,
-			Err: fmt.Errorf("adsfadsfasfdasfd"),
+	if !fail {
+		fexec.AddFakeCmdsNoOutputNoError([]string{
+			"ovn-nbctl --timeout=15 --if-exists get logical_switch_port" + " " + p.portName + " dynamic_addresses addresses",
 		})
+
+		fexec.AddFakeCmdsNoOutputNoError([]string{
+			"ovn-nbctl --timeout=15 --may-exist lsp-add " + p.nodeName + " " + p.portName + " -- lsp-set-addresses " + p.portName + " " + p.podMAC + " " + p.podIP + " -- set logical_switch_port " + p.portName + " external-ids:namespace=" + p.namespace + " external-ids:pod=true -- lsp-set-port-security " + p.portName + " " + p.podMAC + " " + p.podIP,
+		})
+		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
+			Cmd:    "ovn-nbctl --timeout=15 get logical_switch_port " + p.portName + " _uuid",
+			Output: fakeUUID + "\n",
+		})
+		fexec.AddFakeCmdsNoOutputNoError([]string{
+			"ovn-nbctl --timeout=15 --if-exists remove port_group mcastPortGroupDeny ports " + fakeUUID + " -- add port_group mcastPortGroupDeny ports " + fakeUUID,
+		})
+
 	} else {
 		fexec.AddFakeCmdsNoOutputNoError([]string{
-			"ovn-nbctl --timeout=15 lsp-set-port-security " + p.portName + " " + p.podMAC + " " + p.podIP,
+			"ovn-nbctl --timeout=15 --if-exists get logical_switch_port" + " " + p.portName + " dynamic_addresses addresses",
+		})
+		fexec.AddFakeCmd(&ovntest.ExpectedCmd{
+			Cmd: strings.Join([]string{
+				"ovn-nbctl --timeout=15 --may-exist lsp-add " + p.nodeName + " " + p.portName + " -- lsp-set-addresses " + p.portName + " " + p.podMAC + " " + p.podIP + " -- set logical_switch_port " + p.portName + " external-ids:namespace=" + p.namespace + " external-ids:pod=true -- lsp-set-port-security " + p.portName + " " + p.podMAC + " " + p.podIP,
+			}, " "),
+			Err: fmt.Errorf("adsfadsfasfdasfd"),
 		})
 	}
+
 }
 
 func (p pod) addCmdsForNonExistingPod(fexec *ovntest.FakeExec) {
 	p.addCmds(fexec, false)
 }
 
-func (p pod) addCmdsForExistingPod(fexec *ovntest.FakeExec, addPort bool) {
-	// pod setup
-	var addPortCmd string
-	if addPort {
-		addPortCmd = fmt.Sprintf("lsp-add %s %s ", p.nodeName, p.portName)
-	}
-	fexec.AddFakeCmdsNoOutputNoError([]string{
-		fmt.Sprintf("ovn-nbctl --timeout=15 %s-- lsp-set-addresses %s %s %s -- --if-exists clear logical_switch_port %s dynamic_addresses -- set logical_switch_port %s external-ids:namespace=namespace external-ids:pod=true", addPortCmd, p.portName, p.podMAC, p.podIP, p.portName, p.portName),
-	})
-	fexec.AddFakeCmd(&ovntest.ExpectedCmd{
-		Cmd:    "ovn-nbctl --timeout=15 get logical_switch_port " + p.portName + " _uuid",
-		Output: fakeUUID + "\n",
-	})
-	fexec.AddFakeCmdsNoOutputNoError([]string{
-		"ovn-nbctl --timeout=15 lsp-set-port-security " + p.portName + " " + p.podMAC + " " + p.podIP,
-	})
-}
-
-func (p pod) addCmdsForNonExistingFailedPod(fexec *ovntest.FakeExec) {
-	p.addCmds(fexec, true)
-}
-
 func (p pod) delCmds(fexec *ovntest.FakeExec) {
-	// pod's logical switch port is removed
 	fexec.AddFakeCmdsNoOutputNoError([]string{
-		"ovn-nbctl --timeout=15 --if-exists lsp-del " + p.portName,
+		"ovn-nbctl --timeout=15 --if-exists remove port_group mcastPortGroupDeny ports " + fakeUUID,
 	})
 }
 
-func (p pod) delFromNamespaceCmds(fexec *ovntest.FakeExec, pod pod, isMulticastEnabled bool) {
-	fexec.AddFakeCmdsNoOutputNoError([]string{
-		fmt.Sprintf(`ovn-nbctl --timeout=15 remove address_set %s addresses "%s"`, hashedAddressSet(pod.namespace), pod.podIP),
-	})
-	if isMulticastEnabled {
-		fexec.AddFakeCmdsNoOutputNoError([]string{
-			"ovn-nbctl --timeout=15 --if-exists remove port_group mcastPortGroupDeny ports " + fakeUUID,
-		})
-	}
-}
-
 func (p pod) addPodDenyMcast(fexec *ovntest.FakeExec) {
 	fexec.AddFakeCmdsNoOutputNoError([]string{
 		"ovn-nbctl --timeout=15 --if-exists remove port_group mcastPortGroupDeny ports " + fakeUUID + " -- add port_group mcastPortGroupDeny ports " + fakeUUID,
 	})
 }
 
-func (p pod) delPodDenyMcast(fexec *ovntest.FakeExec) {
-	fexec.AddFakeCmdsNoOutputNoError([]string{
-		"ovn-nbctl --timeout=15 --if-exists remove port_group mcastPortGroupDeny ports " + fakeUUID,
-	})
-}
-
 var _ = Describe("OVN Pod Operations", func() {
 	var (
 		app     *cli.App
@@ -198,6 +210,7 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles an existing pod", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				// Setup an unassigned pod, perform an update later on which assigns it.
 				t := newTPod(
 					"",
@@ -205,9 +218,9 @@ var _ = Describe("OVN Pod Operations", func() {
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
@@ -215,14 +228,22 @@ var _ = Describe("OVN Pod Operations", func() {
 					Output: "\n",
 				})
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				fakeOvn.controller.WatchPods()
+				fakeOvn.controller.WatchNamespaces()
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
+				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(context.TODO(), t.podName, metav1.GetOptions{})
 				Expect(err).NotTo(HaveOccurred())
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
 
@@ -233,19 +254,13 @@ var _ = Describe("OVN Pod Operations", func() {
 				t.nodeName = "node1"
 				t.portName = t.namespace + "_" + t.podName
 				t.populateLogicalSwitchCache(fakeOvn)
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				_, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Update(newPod(t.namespace, t.podName, t.nodeName, t.podIP))
+				_, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Update(context.TODO(), newPod(t.namespace, t.podName, t.nodeName, t.podIP), metav1.UpdateOptions{})
 				Expect(err).NotTo(HaveOccurred())
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 
 				return nil
 			}
@@ -256,43 +271,45 @@ var _ = Describe("OVN Pod Operations", func() {
 
 		It("reconciles a new pod", func() {
 			app.Action = func(ctx *cli.Context) error {
-
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				t.baseCmds(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{},
-				})
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
+					},
+					&v1.PodList{
+						Items: []v1.Pod{},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, _ := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
+				pod, _ := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(context.TODO(), t.podName, metav1.GetOptions{})
 				Expect(pod).To(BeNil())
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
 
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				_, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Create(newPod(t.namespace, t.podName, t.nodeName, t.podIP))
+				_, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Create(context.TODO(), newPod(t.namespace, t.podName, t.nodeName, t.podIP), metav1.CreateOptions{})
 				Expect(err).NotTo(HaveOccurred())
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 
 				return nil
 			}
@@ -304,6 +321,7 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles a deleted pod", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				// Setup an assigned pod
 				t := newTPod(
 					"node1",
@@ -311,43 +329,44 @@ var _ = Describe("OVN Pod Operations", func() {
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
 				// Delete it
-				t.delPodDenyMcast(fExec)
 				t.delCmds(fExec)
 
-				err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Delete(t.podName, metav1.NewDeleteOptions(0))
+				err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Delete(context.TODO(), t.podName, *metav1.NewDeleteOptions(0))
 				Expect(err).NotTo(HaveOccurred())
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
+				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(context.TODO(), t.podName, metav1.GetOptions{})
 				Expect(err).To(HaveOccurred())
 				Expect(pod).To(BeNil())
 
@@ -361,6 +380,7 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("retries a failed pod Add on Update", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				// Setup an unassigned pod, perform an update later on which assigns it.
 				t := newTPod(
 					"node1",
@@ -368,40 +388,57 @@ var _ = Describe("OVN Pod Operations", func() {
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
-				t.addCmdsForNonExistingFailedPod(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				mockAddNBDBError(ovntest.LogicalSwitchPortType, t.portName,
+					ovntest.LogicalSwitchPortExternalId,
+					fmt.Errorf("injected dummy port external_ids set error"),
+					fakeOvn.ovnNBClient)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
-
+				// allow pod retry from update annotation to fail
+				time.Sleep(2 * time.Second)
+				mockDelNBDBError(ovntest.LogicalSwitchPortType, t.portName,
+					ovntest.LogicalSwitchPortExternalId,
+					fakeOvn.ovnNBClient)
+				patch := struct {
+					Metadata map[string]interface{} `json:"metadata"`
+				}{
+					Metadata: map[string]interface{}{
+						"annotations": map[string]string{"dummy": "data"},
+					},
+				}
+				patchData, err := json.Marshal(&patch)
+				Expect(err).NotTo(HaveOccurred())
+				// trigger update event
+				_, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Patch(context.TODO(), t.podName, types.MergePatchType, patchData, metav1.PatchOptions{})
+				Expect(err).NotTo(HaveOccurred())
 				// Pod creation should be retried on Update event
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
-				_, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Update(newPod(t.namespace, t.podName, t.nodeName, t.podIP))
-				Expect(err).NotTo(HaveOccurred())
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
-
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
-
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				return nil
 			}
 
@@ -415,15 +452,16 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles an existing pod", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
@@ -433,31 +471,32 @@ var _ = Describe("OVN Pod Operations", func() {
 				fExec.AddFakeCmdsNoOutputNoError([]string{
 					"ovn-nbctl --timeout=15 --if-exists lsp-del " + t.portName,
 				})
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
+				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(context.TODO(), t.podName, metav1.GetOptions{})
 				Expect(err).NotTo(HaveOccurred())
 
 				_, ok := pod.Annotations[util.OvnPodAnnotationName]
 				Expect(ok).To(BeFalse())
 
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 
 				return nil
 			}
@@ -469,29 +508,30 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles a deleted pod", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: t.portName + "\n",
 				})
-
-				t.delCmds(fExec)
+				fExec.AddFakeCmdsNoOutputNoError([]string{
+					"ovn-nbctl --timeout=15 --if-exists lsp-del " + t.portName,
+				})
 
 				fakeOvn.start(ctx)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
-
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
-
 				return nil
 			}
 
@@ -502,39 +542,41 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles a new pod", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
-				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 
 				return nil
 			}
@@ -549,66 +591,58 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles an existing pod without an existing logical switch port", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
-
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
-
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				// Simulate an OVN restart with a new IP assignment and verify that the pod annotation is updated.
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
 				t.populateLogicalSwitchCache(fakeOvn)
-				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
-					Cmd:    "ovn-nbctl --timeout=15 --if-exists get logical_switch_port namespace_myPod _uuid",
-					Output: "\n",
-				})
-				t.addCmdsForExistingPod(fExec, true)
 				t.addPodDenyMcast(fExec)
 
 				fakeOvn.restart()
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
-				// Check that pod annotations have been re-written to correct values
-				podAnnotation, ok = pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
-
 				return nil
 			}
 
@@ -619,66 +653,59 @@ var _ = Describe("OVN Pod Operations", func() {
 		It("reconciles an existing pod with an existing logical switch port", func() {
 			app.Action = func(ctx *cli.Context) error {
 
+				namespaceT := *newNamespace("namespace1")
 				t := newTPod(
 					"node1",
 					"10.128.1.0/24",
 					"10.128.1.2",
 					"10.128.1.1",
 					"myPod",
-					"10.128.1.4",
-					"11:22:33:44:55:66",
-					"namespace",
+					"10.128.1.3",
+					"0a:58:0a:80:01:03",
+					namespaceT.Name,
 				)
 
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
-				t.addCmdsForNonExistingPod(fExec)
 				t.addPodDenyMcast(fExec)
 
-				fakeOvn.start(ctx, &v1.PodList{
-					Items: []v1.Pod{
-						*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
 					},
-				})
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(t.namespace, t.podName, t.nodeName, t.podIP),
+						},
+					},
+				)
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err := fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				Expect(fExec.CalledMatchesExpected()).To(BeTrue(), fExec.ErrorDesc)
 
-				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
-
 				// Simulate an OVN restart with a new IP assignment and verify that the pod annotation is updated.
 				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
 					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
 					Output: "\n",
 				})
 				t.populateLogicalSwitchCache(fakeOvn)
-				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
-					Cmd:    "ovn-nbctl --timeout=15 --if-exists get logical_switch_port namespace_myPod _uuid",
-					Output: fakeUUID + "\n",
-				})
-				t.addCmdsForExistingPod(fExec, false)
 				t.addPodDenyMcast(fExec)
 
 				fakeOvn.restart()
 				t.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
 				fakeOvn.controller.WatchPods()
 
-				pod, err = fakeOvn.fakeClient.CoreV1().Pods(t.namespace).Get(t.podName, metav1.GetOptions{})
-				Expect(err).NotTo(HaveOccurred())
+				Eventually(func() string { return getPodAnnotations(fakeOvn.fakeClient, t.namespace, t.podName) }, 2).Should(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
 				Eventually(fExec.CalledMatchesExpected).Should(BeTrue(), fExec.ErrorDesc)
 
-				// Check that pod annotations have been re-written to correct values
-				podAnnotation, ok = pod.Annotations[util.OvnPodAnnotationName]
-				Expect(ok).To(BeTrue())
-				Expect(podAnnotation).To(MatchJSON(`{"default": {"ip_addresses":["` + t.podIP + `/24"], "mac_address":"` + t.podMAC + `", "gateway_ips": ["` + t.nodeGWIP + `"], "ip_address":"` + t.podIP + `/24", "gateway_ip": "` + t.nodeGWIP + `"}}`))
-
 				return nil
 			}
 
@@ -686,4 +713,94 @@ var _ = Describe("OVN Pod Operations", func() {
 			Expect(err).NotTo(HaveOccurred())
 		})
 	})
+
+	Context("with hybrid overlay gw mode", func() {
+		It("resets the hybrid annotations on update", func() {
+			app.Action = func(ctx *cli.Context) error {
+				namespaceT := *newNamespace("namespace1")
+				namespaceT.Annotations[hotypes.HybridOverlayVTEP] = "1.1.1.1"
+				namespaceT.Annotations[hotypes.HybridOverlayExternalGw] = "2.2.2.2"
+				tP := newTPod(
+					"node1",
+					"10.128.1.0/24",
+					"10.128.1.2",
+					"10.128.1.3",
+					"myPod",
+					"10.128.1.4",
+					"0a:58:0a:80:01:04",
+					namespaceT.Name,
+				)
+
+				fakeOvn.start(ctx,
+					&v1.NamespaceList{
+						Items: []v1.Namespace{
+							namespaceT,
+						},
+					},
+					&v1.PodList{
+						Items: []v1.Pod{
+							*newPod(namespaceT.Name, tP.podName, tP.nodeName, tP.podIP),
+						},
+					},
+				)
+				podMAC := ovntest.MustParseMAC(tP.podMAC)
+				fakeOvn.controller.logicalPortCache.add(tP.nodeName, tP.portName, fakeUUID, podMAC, []*net.IPNet{ovntest.MustParseIPNet(tP.nodeSubnet)})
+				fExec.AddFakeCmd(&ovntest.ExpectedCmd{
+					Cmd:    "ovn-nbctl --timeout=15 --data=bare --no-heading --columns=name find logical_switch_port external_ids:pod=true",
+					Output: "\n",
+				})
+				tP.addPodDenyMcast(fExec)
+				tP.populateLogicalSwitchCache(fakeOvn)
+				fakeOvn.controller.WatchNamespaces()
+				fakeOvn.controller.WatchPods()
+				_, err := fakeOvn.fakeClient.CoreV1().Namespaces().Get(context.TODO(), namespaceT.Name, metav1.GetOptions{})
+				Expect(err).NotTo(HaveOccurred())
+				pod, err := fakeOvn.fakeClient.CoreV1().Pods(tP.namespace).Get(context.TODO(), tP.podName, metav1.GetOptions{})
+				Expect(err).NotTo(HaveOccurred())
+				podAnnotation, ok := pod.Annotations[util.OvnPodAnnotationName]
+				Expect(ok).To(BeTrue())
+				Expect(podAnnotation).To(ContainSubstring(`"gateway_ip":"10.128.1.3"`))
+				// Update namespace to remove annotation
+				namespaceT.Annotations = nil
+				_, err = fakeOvn.fakeClient.CoreV1().Namespaces().Update(context.TODO(), &namespaceT, metav1.UpdateOptions{})
+				Expect(err).NotTo(HaveOccurred())
+				Eventually(func() map[string]string {
+					updatedNs, err := fakeOvn.fakeClient.CoreV1().Namespaces().Get(context.TODO(), namespaceT.Name, metav1.GetOptions{})
+					if err != nil {
+						return map[string]string{"ns": "error"}
+					}
+					return updatedNs.Annotations
+				}).Should(BeEmpty())
+				time.Sleep(time.Second)
+				// Create new pod
+				tP = newTPod(
+					"node1",
+					"10.128.1.0/24",
+					"10.128.1.2",
+					"10.128.1.1",
+					"myPod2",
+					"10.128.1.5",
+					"0a:58:0a:80:01:05",
+					namespaceT.Name,
+				)
+				tP.addPodDenyMcast(fExec)
+				pod, err = fakeOvn.fakeClient.CoreV1().Pods(tP.namespace).Create(context.TODO(), newPod(namespaceT.Name, tP.podName, tP.nodeName, tP.podIP), metav1.CreateOptions{})
+				Expect(err).NotTo(HaveOccurred())
+				var annotation string
+				Eventually(func() string {
+					pod, err = fakeOvn.fakeClient.CoreV1().Pods(tP.namespace).Get(context.TODO(), tP.podName, metav1.GetOptions{})
+					annotation = pod.Annotations[util.OvnPodAnnotationName]
+					return pod.Annotations[util.OvnPodAnnotationName]
+				}).ShouldNot(BeEmpty())
+				Expect(annotation).To(ContainSubstring(`"gateway_ip":"10.128.1.1"`))
+				return nil
+			}
+
+			err := app.Run([]string{
+				app.Name,
+				"-enable-hybrid-overlay",
+			})
+			Expect(err).NotTo(HaveOccurred())
+		})
+	})
 })
